@@ -9,6 +9,7 @@ import itertools
 from copy import deepcopy
 from collections import defaultdict
 from tigramite.models import Models
+import struct
 
 class CausalEffects():
     r"""Causal effect estimation.
@@ -126,16 +127,19 @@ class CausalEffects():
         # and hidden variables
         # (self.graph, self.graph_type, 
         #  self.tau_max, self.hidden_variables) = 
+
         self._construct_graph(graph=graph, graph_type=graph_type,
                               hidden_variables=hidden_variables)
 
         # print(self.graph.shape)
         self._check_graph(self.graph)
 
-        anc_Y = self._get_ancestors(Y)
+        self.ancX = self._get_ancestors(X)
+        self.ancY = self._get_ancestors(Y)
+        self.ancS = self._get_ancestors(S)
 
         # If X is not in anc(Y), then no causal link exists
-        if anc_Y.intersection(set(X)) == set():
+        if self.ancY.intersection(set(X)) == set():
             self.no_causal_path = True
             if self.verbosity > 0:
                 print("No causal path from X to Y exists.")
@@ -169,21 +173,24 @@ class CausalEffects():
         if check_SM_overlap and len(self.S.intersection(self.M)) > 0:
             raise ValueError("Conditions S overlap with mediators M!")
 
-        descendants = self._get_descendants(self.Y.union(self.M))
-        
+        self.desX = self._get_descendants(self.X)
+        self.desY = self._get_descendants(self.Y)
+        self.desM = self._get_descendants(self.M)
+        self.descendants = self.desY.union(self.desM)
+
         # Define forb as X and descendants of YM
-        self.forbidden_nodes = descendants.union(self.X)  #.union(S)
+        self.forbidden_nodes = self.descendants.union(self.X)  #.union(S)
 
         # Define valid ancestors
-        self.vancs = self._get_ancestors(list(self.X.union(self.Y).union(self.S))) - self.forbidden_nodes
+        self.vancs = self.ancX.union(self.ancY).union(self.ancS) - self.forbidden_nodes
 
-        if len(self.S.intersection(self._get_descendants(self.X))) > 0:
-            if self.verbosity > 0:
+        if self.verbosity > 0:
+            if len(self.S.intersection(self.desX)) > 0:
                 print("Warning: Potentially outside assumptions: Conditions S overlap with des(X)")
 
         # Here only check if S overlaps with des(Y), leave the option that S
         # contains variables in des(M) to the user
-        if len(self.S.intersection(self._get_descendants(self.Y))) > 0:
+        if len(self.S.intersection(self.desY)) > 0:
             raise ValueError("Not identifiable: Conditions S overlap with des(Y)")
 
         if self.verbosity > 0:
@@ -264,6 +271,8 @@ class CausalEffects():
 
             stat_graph = deepcopy(graph)
 
+            allowed_edges = ["-->", "<--"]
+
             # Construct tsg_graph
             graph = np.zeros((self.N, self.N, self.tau_max + 1, self.tau_max + 1), dtype='<U3')
             graph[:] = ""
@@ -288,6 +297,16 @@ class CausalEffects():
                         if stat_graph[i, j, tau] == '-->':
                             graph[i, j, taui, tauj] = "-->" 
                             graph[j, i, tauj, taui] = "<--" 
+                        elif stat_graph[i, j, tau] == '<--':
+                            pass
+                        elif stat_graph[i, j, tau] == '':
+                            pass
+                        else:
+                            edge = stat_graph[i, j, tau]
+                            raise ValueError("Invalid graph edge %s. " %(edge) +
+                                 "For graph_type = %s only %s are allowed." %(graph_type, str(allowed_edges)))
+      
+
 
                         # elif stat_graph[i, j, tau] == '<--':
                         #     graph[i, j, taui, tauj] = "<--"
@@ -345,16 +364,16 @@ class CausalEffects():
         oldX = self.X.copy()
         oldY = self.Y.copy()
 
-        anc_Y = self._get_ancestors(self.Y)
-        anc_S = self._get_ancestors(self.S)
+        # anc_Y = self._get_ancestors(self.Y)
+        # anc_S = self._get_ancestors(self.S)
 
         # Remove first from X those nodes with no causal path to Y or S
-        X = set([x for x in self.X if x in anc_Y.union(anc_S)])
+        X = set([x for x in self.X if x in self.ancY.union(self.ancS)])
         
         # Remove from Y those nodes with no causal path from X
-        des_X = self._get_descendants(X)
+        # des_X = self._get_descendants(X)
 
-        Y = set([y for y in self.Y if y in des_X])
+        Y = set([y for y in self.Y if y in self.desX])
 
         # Also require that all x in X have proper path to Y or S,
         # that is, the first link goes out of x 
@@ -596,27 +615,33 @@ class CausalEffects():
         # Find adjacencies going forward/contemp
         for k, lag_ik in zip(*np.where(graph[i,:,lag_i,:])):
             # print((k, lag_ik), graph[i,k,lag_i,lag_ik]) 
-            matches = [self._match_link(patt, graph[i,k,lag_i,lag_ik]) for patt in patterns]
-            if np.any(matches):
-                match = (k, -lag_ik)
-                if match not in exclude:
-                    if return_link:
-                        adj.append((graph[i,k,lag_i,lag_ik], match))
-                    else:
-                        adj.append(match)
+            # matches = [self._match_link(patt, graph[i,k,lag_i,lag_ik]) for patt in patterns]
+            # if np.any(matches):
+            for patt in patterns:
+                if self._match_link(patt, graph[i,k,lag_i,lag_ik]):
+                    match = (k, -lag_ik)
+                    if match not in exclude:
+                        if return_link:
+                            adj.append((graph[i,k,lag_i,lag_ik], match))
+                        else:
+                            adj.append(match)
+                    break
 
         
         # Find adjacencies going backward/contemp
         for k, lag_ki in zip(*np.where(graph[:,i,:,lag_i])):  
             # print((k, lag_ki), graph[k,i,lag_ki,lag_i]) 
-            matches = [self._match_link(self._reverse_link(patt), graph[k,i,lag_ki,lag_i]) for patt in patterns]
-            if np.any(matches):
-                match = (k, -lag_ki)
-                if match not in exclude:
-                    if return_link:
-                        adj.append((self._reverse_link(graph[k,i,lag_ki,lag_i]), match))
-                    else:
-                        adj.append(match)
+            # matches = [self._match_link(self._reverse_link(patt), graph[k,i,lag_ki,lag_i]) for patt in patterns]
+            # if np.any(matches):
+            for patt in patterns:
+                if self._match_link(self._reverse_link(patt), graph[k,i,lag_ki,lag_i]):
+                    match = (k, -lag_ki)
+                    if match not in exclude:
+                        if return_link:
+                            adj.append((self._reverse_link(graph[k,i,lag_ki,lag_i]), match))
+                        else:
+                            adj.append(match)
+                    break
      
         adj = list(set(adj))
         return adj
@@ -634,7 +659,7 @@ class CausalEffects():
         return ((tauij >= 0 and self._match_link(pattern_ij, graph[i, j, tauij])) or
                (tauij < 0 and self._match_link(self._reverse_link(pattern_ij), graph[j, i, abs(tauij)])))
 
-
+    # @profile
     def _get_children(self, varlag):
         """Returns set of children (varlag --> ...) for (lagged) varlag."""
         if self.possible:
@@ -745,6 +770,7 @@ class CausalEffects():
 
         return descendants
 
+    # @profile
     def _get_descendants(self, W):
         """Get descendants of nodes in W up to time t.
         
@@ -1079,6 +1105,7 @@ class CausalEffects():
 
         return aux_graph
 
+    # @profile
     def _check_path(self, 
         # graph, 
         start, end,
@@ -1280,7 +1307,7 @@ class CausalEffects():
         # print("Separated")
         return False
 
-
+    # @profile
     def get_optimal_set(self, 
         alternative_conditions=None,
         minimize=False,
@@ -1314,9 +1341,16 @@ class CausalEffects():
             vancs = self.vancs.copy()
         else:
             S = alternative_conditions
-            vancs = self._get_ancestors(list(self.X.union(self.Y).union(S))) - self.forbidden_nodes
+            newancS = self._get_ancestors(S)
+            self.vancs = self.ancX.union(self.ancY).union(newancS) - self.forbidden_nodes
 
-        descendants = self._get_descendants(self.Y.union(self.M))
+            # vancs = self._get_ancestors(list(self.X.union(self.Y).union(S))) - self.forbidden_nodes
+
+        # descendants = self._get_descendants(self.Y.union(self.M))
+
+        # Sufficient condition for non-identifiability
+        if len(self.X.intersection(self.descendants)) > 0:
+            return False  # raise ValueError("Not identifiable: Overlap between X and des(M)")
 
         ##
         ## Construct O-set
@@ -1405,13 +1439,6 @@ class CausalEffects():
 
         Oset_S = Oset.union(S)
 
-        # For singleton X the validity is already checked in the
-        # if-statements of the construction algorithm, but for 
-        # multivariate X there might be further cases... Hence,
-        # we here explicitely check validity
-        if self._check_validity(list(Oset_S)) is False:
-            return False
-
         if return_separate_sets:
             return parents, colliders, collider_parents, S
         else:
@@ -1769,7 +1796,7 @@ class CausalEffects():
 
         return all_causal_paths
 
-
+    # @profile
     def fit_total_effect(self,
         dataframe, 
         estimator,
@@ -1826,6 +1853,9 @@ class CausalEffects():
             if self._check_validity(adjustment_set) is False:
                 raise ValueError("Chosen adjustment_set is not valid.")
 
+        if adjustment_set is False:
+            raise ValueError("Causal effect not identifiable via adjustment.")
+
         self.adjustment_set = adjustment_set
 
         # Fit model of Y on X and Z (and conditions)
@@ -1847,10 +1877,12 @@ class CausalEffects():
 
         return self
 
+    # @profile
     def predict_total_effect(self, 
         intervention_data, 
         conditions_data=None,
         pred_params=None,
+        return_further_pred_results=False,
         ):
         """Predict effect of intervention with fitted model.
 
@@ -1864,17 +1896,20 @@ class CausalEffects():
             Numpy array of shape (time, len(S)) that contains the S=s values.
         pred_params : dict, optional
             Optional parameters passed on to sklearn prediction function.
+        return_further_pred_results : bool, optional (default: False)
+            In case the predictor class returns more than just the expected value,
+            the entire results can be returned.
 
         Returns
         -------
         Results from prediction: an array of shape  (time, len(Y)).
         """
 
-        if intervention_data.shape[1] != len(self.X):
+        if intervention_data.shape[1] != len(self.listX):
             raise ValueError("intervention_data.shape[1] must be len(X).")
 
         if conditions_data is not None:
-            if conditions_data.shape[1] != len(self.S):
+            if conditions_data.shape[1] != len(self.listS):
                 raise ValueError("conditions_data.shape[1] must be len(S).")
             if conditions_data.shape[0] != intervention_data.shape[0]:
                 raise ValueError("conditions_data.shape[0] must match intervention_data.shape[0].")
@@ -1882,15 +1917,17 @@ class CausalEffects():
         if self.no_causal_path:
             if self.verbosity > 0:
                 print("No causal path from X to Y exists.")
-            return np.zeros((len(intervention_data), len(self.Y)))
+            return np.zeros((len(intervention_data), len(self.listY)))
 
         effect = self.model.get_general_prediction(
             intervention_data=intervention_data,
             conditions_data=conditions_data,
-            pred_params=pred_params) 
+            pred_params=pred_params,
+            return_further_pred_results=return_further_pred_results) 
 
         return effect
 
+    # @profile
     def fit_wright_effect(self,
         dataframe, 
         mediation=None,
@@ -1949,7 +1986,7 @@ class CausalEffects():
                         mask_type=mask_type,
                         verbosity=self.verbosity)
 
-        mediators = self.get_mediators(start=self.X, end=self.Y)
+        mediators = self.M  # self.get_mediators(start=self.X, end=self.Y)
 
         if mediation == 'direct':
             causal_paths = {}         
@@ -2042,6 +2079,8 @@ class CausalEffects():
 
                 effect[(x, y)] += effect_here
                
+        # Make fitted coefficients available as attribute
+        self.coeffs = coeffs
 
         # Modify and overwrite variables in self.model
         self.model.Y = self.listY
@@ -2067,7 +2106,7 @@ class CausalEffects():
         self.model.fit_results = fit_results
         return self
 
-    
+    # @profile 
     def predict_wright_effect(self, 
         intervention_data, 
         pred_params=None,
@@ -2095,11 +2134,11 @@ class CausalEffects():
 
         intervention_T, lenX = intervention_data.shape
 
-        lenY = len(self.Y)
+        lenY = len(self.listY)
 
         predicted_array = np.zeros((intervention_T, lenY))
         pred_dict = {}
-        for iy, y in enumerate(self.Y):
+        for iy, y in enumerate(self.listY):
             # Print message
             if self.verbosity > 1:
                 print("\n## Predicting target %s" % str(y))
@@ -2129,4 +2168,158 @@ class CausalEffects():
                 X=predictor_array, **pred_params)
                 predicted_array[index, iy] = predicted_vals.mean()
 
-        return predicted_array
\ No newline at end of file
+        return predicted_array
+
+    @staticmethod
+    def get_graph_from_dict(links, tau_max=None):
+        """Helper function to convert dictionary of links to graph array format.
+
+        Parameters
+        ---------
+        links : dict
+            Dictionary of form {0:[((0, -1), coeff, func), ...], 1:[...], ...}.
+            Also format {0:[(0, -1), ...], 1:[...], ...} is allowed.
+        tau_max : int or None
+            Maximum lag. If None, the maximum lag in links is used.
+
+        Returns
+        -------
+        graph : array of shape (N, N, tau_max+1)
+            Matrix format of graph with 1 for true links and 0 else.
+        """
+
+        def _get_minmax_lag(links):
+            """Helper function to retrieve tau_min and tau_max from links.
+            """
+
+            N = len(links)
+
+            # Get maximum time lag
+            min_lag = np.inf
+            max_lag = 0
+            for j in range(N):
+                for link_props in links[j]:
+                    if len(link_props) > 2:
+                        var, lag = link_props[0]
+                        coeff = link_props[1]
+                        # func = link_props[2]
+                        if coeff != 0.:
+                            min_lag = min(min_lag, abs(lag))
+                            max_lag = max(max_lag, abs(lag))
+                    else:
+                        var, lag = link_props
+                        min_lag = min(min_lag, abs(lag))
+                        max_lag = max(max_lag, abs(lag))   
+
+            return min_lag, max_lag
+
+        N = len(links)
+
+        # Get maximum time lag
+        min_lag, max_lag = _get_minmax_lag(links)
+
+        # Set maximum lag
+        if tau_max is None:
+            tau_max = max_lag
+        else:
+            if max_lag > tau_max:
+                raise ValueError("tau_max is smaller than maximum lag = %d "
+                                 "found in links, use tau_max=None or larger "
+                                 "value" % max_lag)
+
+        graph = np.zeros((N, N, tau_max + 1), dtype='<U3')
+        for j in links.keys():
+            for link_props in links[j]:
+                if len(link_props) > 2:
+                    var, lag = link_props[0]
+                    coeff = link_props[1]
+                    if coeff != 0.:
+                        graph[var, j, abs(lag)] = "-->"
+                        if lag == 0:
+                            graph[j, var, 0] = "<--"
+                else:
+                    var, lag = link_props
+                    graph[var, j, abs(lag)] = "-->"
+                    if lag == 0:
+                        graph[j, var, 0] = "<--"
+
+        return graph
+
+if __name__ == '__main__':
+    
+    # Consider some toy data
+    import tigramite
+    import tigramite.toymodels.structural_causal_processes as toys
+    import tigramite.data_processing as pp
+
+    import sklearn
+    from sklearn.linear_model import LinearRegression
+
+    T = 1000
+    def lin_f(x): return x
+    auto_coeff = 0.3
+    coeff = 2.
+    links = {
+            0: [((0, -1), auto_coeff, lin_f)], 
+            1: [((1, -1), auto_coeff, lin_f), ((0, -1), coeff, lin_f)], 
+            2: [((2, -1), auto_coeff, lin_f), ((1, 0), coeff, lin_f)],
+            }
+    data, nonstat = toys.structural_causal_process(links, T=T, 
+                                noises=None, seed=7)
+
+    # Create some missing values
+    data[-10:,:] = 999.
+    var_names = range(5)
+    dataframe = pp.DataFrame(data, var_names=var_names,
+     missing_flag=999.) 
+
+
+    # Construct expert knowledge graph from links here 
+    links = {0: [(0, -1)],
+             1: [(1, -1), (0, -1)],
+             2: [(2, -1), (1, 0),],
+             }
+    # Use staticmethod to get graph
+    graph = CausalEffects.get_graph_from_dict(links, tau_max=None)
+    
+    # We are interested in lagged total effect of X on Y
+    X = [(0, -1)]
+    Y = [(2, 0)]
+
+    # Initialize class as `stationary_dag`
+    causal_effects = CausalEffects(graph, graph_type='stationary_dag', 
+                                X=X, Y=Y, S=None, 
+                                hidden_variables=None, 
+                                verbosity=5)
+
+    print(data)
+    # Optimal adjustment set (is used by default)
+    # print(causal_effects.get_optimal_set())
+
+    # # Fit causal effect model from observational data
+    # causal_effects.fit_total_effect(
+    #     dataframe=dataframe, 
+    #     # mask_type='y',
+    #     estimator=LinearRegression(),
+    #     )
+
+    # # Predict effect of interventions do(X=0.), ..., do(X=1.) in one go
+    # dox_vals = np.linspace(0., 1., 5)
+    # intervention_data = dox_vals.reshape(len(dox_vals), len(X))
+    # pred_Y = causal_effects.predict_total_effect( 
+    #         intervention_data=intervention_data)
+    # print(pred_Y)
+
+    # Fit causal effect model from observational data
+    causal_effects.fit_wright_effect(
+        dataframe=dataframe, 
+        # mask_type='y',
+        # estimator=LinearRegression(),
+        )
+
+    # Predict effect of interventions do(X=0.), ..., do(X=1.) in one go
+    dox_vals = np.linspace(0., 1., 5)
+    intervention_data = dox_vals.reshape(len(dox_vals), len(X))
+    pred_Y = causal_effects.predict_wright_effect( 
+            intervention_data=intervention_data)
+    print(pred_Y)
\ No newline at end of file
