@@ -17,8 +17,6 @@ from numba import jit
 class DataFrame():
     """Data object containing time series array and optional mask.
 
-    Alternatively, a panda dataframe can be used.
-
     Parameters
     ----------
     data : array-like
@@ -42,22 +40,28 @@ class DataFrame():
         passed, variables are enumerated as [0, 1, ...]
     datatime : array-like, optional (default: None)
         Timelabel array. If None, range(T) is used.
+    remove_missing_upto_maxlag : bool, optional (default: False)
+        Whether to remove not only missing samples, but also all neighboring
+        samples up to max_lag (as given by cut_off in construct_array).
     """
     def __init__(self, data, mask=None, missing_flag=None, var_names=None,
-        datatime=None):
+        datatime=None, remove_missing_upto_maxlag=False):
 
-        self.values = data
+        self.values = data.copy()
         self.mask = mask
         self.missing_flag = missing_flag
         if self.missing_flag is not None:
             self.values[self.values == self.missing_flag] = np.nan
+        self.remove_missing_upto_maxlag = remove_missing_upto_maxlag
         T, N = data.shape
         # Set the variable names
         self.var_names = var_names
         # Set the default variable names if none are set
         if self.var_names is None:
             self.var_names = {i: i for i in range(N)}
-
+        else:
+            if len(self.var_names) != N:
+                raise ValueError("len(var_names) != data.shape[1].")
         # Set datatime
         self.datatime = datatime
         if self.datatime is None:
@@ -114,6 +118,7 @@ class DataFrame():
                                  % str(_use_mask.shape))
 
     def construct_array(self, X, Y, Z, tau_max,
+                        extraZ=None,
                         mask=None,
                         mask_type=None,
                         return_cleaned_xyz=False,
@@ -127,10 +132,10 @@ class DataFrame():
 
         Parameters
         ----------
-        X, Y, Z : list of tuples
+        X, Y, Z, extraZ : list of tuples
             For a dependence measure I(X;Y|Z), X, Y, Z can be multivariate of
             the form [(var1, -lag), (var2, -lag), ...]. At least one varlag in Y 
-            has to be at lag zero.
+            has to be at lag zero. extraZ is only used in CausalEffects class.
         tau_max : int
             Maximum time lag. This may be used to make sure that estimates for
             different lags in X and Z all have the same sample size.
@@ -171,16 +176,20 @@ class DataFrame():
         # Get the length in time and the number of nodes
         T, N = self.values.shape
 
-        # Remove duplicates in X, Y, Z
+        if extraZ is None:
+            extraZ = []
+        # Remove duplicates in X, Y, Z, extraZ
         X = list(OrderedDict.fromkeys(X))
         Y = list(OrderedDict.fromkeys(Y))
         Z = list(OrderedDict.fromkeys(Z))
+        extraZ = list(OrderedDict.fromkeys(extraZ))
 
         # If a node in Z occurs already in X or Y, remove it from Z
         Z = [node for node in Z if (node not in X) and (node not in Y)]
+        extraZ = [node for node in extraZ if (node not in X) and (node not in Y) and (node not in Z)]
 
         # Check that all lags are non-positive and indices are in [0,N-1]
-        XYZ = X + Y + Z
+        XYZ = X + Y + Z + extraZ
         dim = len(XYZ)
 
         # Ensure that XYZ makes sense
@@ -200,9 +209,10 @@ class DataFrame():
         # Setup XYZ identifier
         index_code = {'x' : 0,
                       'y' : 1,
-                      'z' : 2}
+                      'z' : 2,
+                      'e' : 3}
         xyz = np.array([index_code[name]
-                        for var, name in zip([X, Y, Z], ['x', 'y', 'z'])
+                        for var, name in zip([X, Y, Z, extraZ], ['x', 'y', 'z', 'e'])
                         for _ in var])
 
         # Setup and fill array with lagged time series
@@ -218,15 +228,23 @@ class DataFrame():
         # Choose which indices to use
         use_indices = np.ones(time_length, dtype='int')
 
-        # Remove all values that have missing value flag, as well as the time
+        # Remove all values that have missing value flag, and optionally as well the time
         # slices that occur up to max_lag after
         if self.missing_flag is not None:
-            missing_anywhere = np.any(np.isnan(self.values), axis=1)
-            for tau in range(max_lag+1):
+            missing_anywhere = np.array(np.where(np.any(np.isnan(array), axis=0))[0])
+            if self.remove_missing_upto_maxlag:
+                for tau in range(max_lag+1):
+                    if self.bootstrap is None:
+                        delete = missing_anywhere + tau 
+                        delete = delete[delete < time_length]
+                        use_indices[delete] = 0
+                    else:
+                        use_indices[missing_anywhere[self.bootstrap] + tau] = 0
+            else:
                 if self.bootstrap is None:
-                    use_indices[missing_anywhere[tau:T-max_lag+tau]] = 0
+                    use_indices[missing_anywhere] = 0
                 else:
-                    use_indices[missing_anywhere[self.bootstrap - max_lag + tau]] = 0
+                    use_indices[missing_anywhere[self.bootstrap]] = 0
 
         # Use the mask override if needed
         _use_mask = mask
@@ -250,7 +268,7 @@ class DataFrame():
                     array_mask[i, :] = (_use_mask[self.bootstrap + lag, var] == False)
 
             # Iterate over defined mapping from letter index to number index,
-            # i.e. 'x' -> 0, 'y' -> 1, 'z'-> 2
+            # i.e. 'x' -> 0, 'y' -> 1, 'z'-> 2, 'e'-> 3
             for idx, cde in index_code.items():
                 # Check if the letter index is in the mask type
                 if (mask_type is not None) and (idx in mask_type):
@@ -268,7 +286,7 @@ class DataFrame():
 
         # Print information about the constructed array
         if verbosity > 2:
-            self.print_array_info(array, X, Y, Z, self.missing_flag, mask_type)
+            self.print_array_info(array, X, Y, Z, self.missing_flag, mask_type, extraZ)
 
         # Return the array and xyz and optionally (X, Y, Z)
         if return_cleaned_xyz:
@@ -310,7 +328,7 @@ class DataFrame():
         #     raise ValueError("Y-nodes are %s, " % str(Y) +
         #                      "but one of the Y-nodes must have zero lag")
 
-    def print_array_info(self, array, X, Y, Z, missing_flag, mask_type):
+    def print_array_info(self, array, X, Y, Z, missing_flag, mask_type, extraZ=None):
         """
         Print info about the constructed array
 
@@ -318,7 +336,7 @@ class DataFrame():
         ----------
         array : Data array of shape (dim, T)
             Data array.
-        X, Y, Z : list of tuples
+        X, Y, Z, extraZ : list of tuples
             For a dependence measure I(X;Y|Z), Y is of the form [(varY, 0)],
             where var specifies the variable index. X typically is of the form
             [(varX, -tau)] with tau denoting the time lag and Z can be
@@ -333,11 +351,15 @@ class DataFrame():
             measure I(X; Y | Z) the samples should be masked. If None, the mask
             is not used. Explained in tutorial on masking and missing values.
         """
+        if extraZ is None:
+            extraZ = []
         indt = " " * 12
         print(indt + "Constructed array of shape %s from"%str(array.shape) +
               "\n" + indt + "X = %s" % str(X) +
               "\n" + indt + "Y = %s" % str(Y) +
               "\n" + indt + "Z = %s" % str(Z))
+        if extraZ is not None:  
+            print(indt + "extraZ = %s" % str(extraZ))
         if self.mask is not None and mask_type is not None:
             print(indt+"with masked samples in %s removed" % mask_type)
         if self.missing_flag is not None:
@@ -706,6 +728,7 @@ def structural_causal_process(links, T, noises=None,
 
 if __name__ == '__main__':
     
+    from tigramite.toymodels.structural_causal_processes import structural_causal_process
     ## Generate some time series from a structural causal process
     def lin_f(x): return x
     def nonlin_f(x): return (x + 5. * x**2 * np.exp(-x**2 / 20.))
@@ -718,3 +741,5 @@ if __name__ == '__main__':
     data, nonstat = structural_causal_process(links,
      T=100, noises=noises)
     print(data.shape)
+
+    frame = DataFrame(data)
