@@ -6,24 +6,30 @@
 
 from __future__ import print_function
 from copy import deepcopy
-
+import json, warnings, os, pathlib
 import numpy as np
-
+try:
+    from importlib import metadata
+except ImportError:
+    import importlib_metadata as metadata  # python<=3.7
+try:
+    import sklearn
+    import sklearn.linear_model
+    import networkx
+    with open(pathlib.Path(os.path.dirname(__file__)) / '../versions.py', 'r') as vfile:
+        packages = json.loads(vfile.read())['all']
+        packages = dict(map(lambda s: s.split('>='), packages))
+        if metadata.version('scikit-learn') < packages['scikit-learn']:
+            raise Exception('Version mismatch. Installed version of scikit-learn', metadata.version('scikit-learn'),
+                          'Please install scikit-learn>=', packages['scikit-learn'])
+        if metadata.version('networkx') < packages['networkx']:
+            raise Exception('Version mismatch. Installed version of networkx', metadata.version('networkx'),
+                          'Please install networkx>=', packages['networkx'])
+except Exception as e:
+    warnings.warn(str(e))
 from tigramite.data_processing import DataFrame
 from tigramite.pcmci import PCMCI
 
-# try:
-import sklearn
-import sklearn.linear_model
-# except:
-#     print("Could not import sklearn...")
-
-try:
-    import networkx
-except:
-    print("Could not import networkx, LinearMediation plots not possible...")
-
-
 class Models():
     """Base class for time series models.
 
@@ -81,6 +87,7 @@ class Models():
         self.tau_max = None
         self.fit_results = None
 
+    # @profile    
     def get_general_fitted_model(self, 
                 Y, X, Z=None,
                 conditions=None,
@@ -154,7 +161,9 @@ class Models():
             # Construct array of shape (var, time) with first entry being
             # a dummy, second is y followed by joint X and Z (ignore the notation in construct_array)
             array, xyz = \
-                self.dataframe.construct_array(X=self.X, Y=[y] + self.Z, Z=self.conditions,
+                self.dataframe.construct_array(X=self.X, Y=[y], # + self.Z, 
+                                               Z=self.conditions,
+                                               extraZ=self.Z,
                                                tau_max=self.tau_max,
                                                mask_type=self.mask_type,
                                                cut_off=self.cut_off,
@@ -170,8 +179,8 @@ class Models():
             a_model = deepcopy(self.model)
 
             predictor_indices =  list(np.where(xyz==0)[0]) \
-                               + list(np.where(xyz==1)[0][1:]) \
-                               + list(np.where(xyz==2)[0])
+                               + list(np.where(xyz==2)[0]) \
+                               + list(np.where(xyz==3)[0])
             predictor_array = array[predictor_indices, :].T
             # Target is only first entry of Y, ie [y]
             target_array = array[np.where(xyz==1)[0][0], :]
@@ -193,10 +202,12 @@ class Models():
         self.fit_results = fit_results
         return fit_results
 
+    # @profile
     def get_general_prediction(self,
                 intervention_data,
                 conditions_data=None,
                 pred_params=None,
+                return_further_pred_results=False,
                 ):
         r"""Predict effect of intervention with fitted model.
 
@@ -210,7 +221,9 @@ class Models():
             Numpy array of shape (time, len(S)) that contains the S=s values.
         pred_params : dict, optional
             Optional parameters passed on to sklearn prediction function.
-
+        return_further_pred_results : bool, optional (default: False)
+            In case the predictor class returns more than just the expected value,
+            the entire results can be returned.
         Returns
         -------
         Results from prediction.
@@ -234,6 +247,7 @@ class Models():
         predicted_array = np.zeros((intervention_T, lenY))
         pred_dict = {}
         for iy, y in enumerate(self.Y):
+            pred_dict[iy] = {}
             # Print message
             if self.verbosity > 1:
                 print("\n## Predicting target %s" % str(y))
@@ -255,7 +269,7 @@ class Models():
                     conditions_data = a_transform.transform(X=conditions_data)
 
             # Extract observational Z from stored array
-            z_indices = list(np.where(self.fit_results[y]['xyz']==1)[0][1:])
+            z_indices = list(np.where(self.fit_results[y]['xyz']==3)[0])
             z_array = self.fit_results[y]['observation_array'][z_indices, :].T  
             Tobs = len(z_array)              
 
@@ -282,16 +296,28 @@ class Models():
                     #     predicted_vals = a_transform.transform(X=target_array.T).T
                     a_conditional_model = deepcopy(self.conditional_model)
                     
-                    a_conditional_model.fit(X=s_array, y=predicted_vals)
+                    if type(predicted_vals) is tuple:
+                        predicted_vals_here = predicted_vals[0]
+                    else:
+                        predicted_vals_here = predicted_vals
+
+                    a_conditional_model.fit(X=s_array, y=predicted_vals_here)
                     self.fit_results[y]['conditional_model'] = a_conditional_model
 
-                    predicted_array[index, iy] = a_conditional_model.predict(
-                        X=conditions_array, **pred_params).mean()
+                    predicted_vals = a_conditional_model.predict(
+                        X=conditions_array, **pred_params)
 
+                # print(predicted_vals)
+                if type(predicted_vals) is tuple:
+                    predicted_array[index, iy] = predicted_vals[0].mean()
+                    pred_dict[iy][index] = predicted_vals
                 else:
                     predicted_array[index, iy] = predicted_vals.mean()
 
-        return predicted_array
+        if return_further_pred_results:
+            return predicted_array, pred_dict
+        else:
+            return predicted_array
 
 
     def get_fit(self, all_parents,
